﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using DDStoDXT.Types;
using System.Security.Cryptography;

namespace DDStoDXT.IO
{
    public class FileReader : IDisposable
    {
        private bool disposed;
        protected readonly Stream stream;

        // Constant buffer for read operations:
        protected readonly byte[] buffer = new byte[32];

        public string Name
        {
            get; protected set;
        }

        public uint Position
        {
            get
            {
                return (uint)stream.Position;
            }
            set
            {
                stream.Position = value;
            }
        }

        public uint Size
        {
            get
            {
                return (uint)stream.Length;
            }
        }

        public FileReader(Stream stream)
        {
            this.stream = stream;
        }

        public FileReader(string path)
            : this(new FileStream(path, FileMode.Open, FileAccess.Read, FileShare.Read))
        {
            this.Name = Path.GetFileName(path);
        }

        // BYTES:

        public byte ReadU8()
        {
            InternalRead(buffer, 0, 1);
            return buffer[0];
        }

        public sbyte ReadS8()
        {
            InternalRead(buffer, 0, 1);
            return (sbyte)(buffer[0]);
        }

        // LITTLE ENDIAN:

        public ushort ReadU16LE()
        {
            InternalRead(buffer, 0, 2);
            return (ushort)(buffer[0] | buffer[1] << 8);
        }

        public uint ReadU24LE()
        {
            InternalRead(buffer, 0, 3);
            return (uint)(buffer[0] | buffer[1] << 8 | buffer[2] << 16);
        }

        public uint ReadU32LE()
        {
            InternalRead(buffer, 0, 4);
            return (uint)(buffer[0] | buffer[1] << 8 | buffer[2] << 16 | buffer[3] << 24);
        }

        public short ReadS16LE()
        {
            InternalRead(buffer, 0, 2);
            return (short)(buffer[0] | buffer[1] << 8);
        }

        public int ReadS32LE()
        {
            InternalRead(buffer, 0, 4);
            return buffer[0] | buffer[1] << 8 | buffer[2] << 16 | buffer[3] << 24;
        }

        // BIG ENDIAN:

        public ushort ReadU16BE()
        {
            InternalRead(buffer, 0, 2);
            return (ushort)(buffer[0] << 8 | buffer[1]);
        }

        public uint ReadU24BE()
        {
            InternalRead(buffer, 0, 3);
            return (uint)(buffer[0] << 16 | buffer[1] << 8 | buffer[2]);
        }

        public uint ReadU32BE()
        {
            InternalRead(buffer, 0, 4);
            return (uint)(buffer[0] << 24 | buffer[1] << 16 | buffer[2] << 8 | buffer[3]);
        }

        public short ReadS16BE()
        {
            InternalRead(buffer, 0, 2);
            return (short)(buffer[0] << 8 | buffer[1]);
        }

        public int ReadS32BE()
        {
            InternalRead(buffer, 0, 4);
            return buffer[0] << 24 | buffer[1] << 16 | buffer[2] << 8 | buffer[3];
        }

        public uint ReadVariableLength()
        {
            uint result = 0;
            while (true)
            {
                byte b = ReadU8();
                result <<= 7;
                result |= (uint)(b & 0x7F);
                if ((b & 0x80) == 0)
                {
                    return result;
                }
            }
        }

        // STRINGS:

        // TODO: Optimize
        public string ReadStringZ()
        {
            StringBuilder builder = new StringBuilder();
            byte b;
            do
            {
                b = ReadU8();
                if (b != 0)
                {
                    builder.Append((char)b);
                }
            }
            while (b != 0);

            return builder.ToString();
        }

        public string ReadStringZ(byte nameEncryption)
        {
            StringBuilder builder = new StringBuilder();
            byte b;
            do
            {
                b = ReadU8();
                if (b != 0)
                {
                    builder.Append((char)(b ^ nameEncryption));
                }
            }
            while (b != 0);

            return builder.ToString();
        }

        public string ReadString(int length)
        {
            byte[] buff = new byte[length];
            InternalRead(buff, 0, length);
            return Encoding.ASCII.GetString(buff);
        }

        public FourCC ReadFourCC()
        {
            return new FourCC(ReadU32LE());
        }

        public int Read(int count, out byte[] result)
        {
            result = new byte[count];
            int read = InternalRead(result, 0, count);
            return read;
        }

        public int Read(byte[] aBuffer, int offset, int count)
        {
            return InternalRead(aBuffer, offset, count);
        }

        protected virtual int InternalRead(byte[] aBuffer, int index, int count)
        {
            return stream.Read(aBuffer, 0, count);
        }

        public void Close()
        {
            Dispose();
        }

        public void Dispose()
        {
            Dispose(true);
        }

        /// <summary>
        /// Note: This method is *not* for integrity tests, but rather for 
        /// recognition in a fixed set of widely varying files. It only uses
        /// part of the file content along with file size for performance reasons.
        /// </summary>
        /// <returns></returns>
        public string GetRecognitionChecksum()
        {
            byte[] buffer = new byte[1024];
            int bytesRead = stream.Read(buffer, 0, 1024);

            MD5 md5 = MD5.Create();
            byte[] hash = md5.ComputeHash(buffer, 0, bytesRead);
            string checksum = BitConverter.ToString(hash);
            checksum += String.Format("{0:x8}", Size);
            return checksum;
        }

        public void Dispose(bool disposing)
        {
            if (disposed) return;
            if (disposing)
            {
                stream.Close();
                disposed = true;
            }
        }

    }
}
